package com.rits.orm

import com.rits.jdbc.Jdbc
import com.rits.orm.drivers.Driver
import scala.collection.mutable.HashMap
import com.rits.jdbc.JdbcMap

/**
 * @author kostantinos.kougios
 *
 * 13 Jul 2011
 */
class MapperDao(driver: Driver, typeRegistry: TypeRegistry, typeManager: TypeManager) {
	/**
	 * ===================================================================================
	 * Utility methods
	 * ===================================================================================
	 */

	private def isPersisted(o: Any): Boolean = o.isInstanceOf[Persisted]

	private def addToMap(key: String, v: Any, m: scala.collection.mutable.Map[String, List[Any]]) =
		{
			var l = m.getOrElse(key, List[Any]())
			l ::= v
			m(key) = l
			l
		}
	/**
	 * ===================================================================================
	 * CRUD OPERATIONS
	 * ===================================================================================
	 */

	private def insertInner[T, P](o: T, parent: P, parentColumn: ColumnBase, parentKeysAndValues: List[(SimpleColumn, Any)]): T =
		{
			if (o.isInstanceOf[Persisted]) throw new IllegalArgumentException("can't insert an object that is already persisted: " + o);

			val tpe = typeRegistry.typeInfo(o).tpe
			val table = tpe.table

			val modified = ValuesMap.fromEntity(typeManager, tpe, o).toMutableMap
			val modifiedTraversables = new HashMap[String, List[Any]]
			// extra args for foreign keys
			val extraArgs = if (parent != null) {
				// parent of the one-to-many
				val pti = typeRegistry.typeInfo(parent)
				val table = pti.tpe.table
				parentColumn match {
					case otm: OneToMany =>
						val foreignKeyColumns = otm.foreignColumns
						val foreignKeys = parentKeysAndValues.map(_._2)
						if (foreignKeys.size != foreignKeyColumns.size) throw new IllegalArgumentException("mappings of one-to-many from " + parent + " to " + o + " is invalid. Number of FK columns doesn't match primary keys. columns: " + foreignKeyColumns + " , primary key values " + foreignKeys);
						foreignKeyColumns.zip(foreignKeys)
				}
			} else List()

			// arguments
			val args = table.toListOfColumnAndValueTuples(table.simpleTypeNotAutoGeneratedColumns, o) ::: extraArgs

			if (!args.isEmpty) {
				val ur = driver.doInsert(tpe, args)

				table.simpleTypeAutoGeneratedColumns.foreach { c =>
					modified(c.columnName) = ur.keys.get(c.columnName).get
				}
			}

			val newKeyValues = table.primaryKeys.map(c => modified(c.columnName))
			lazy val newKeyColumnAndValues = table.primaryKeys.map(_.column) zip newKeyValues

			// one to many
			table.oneToManyColumnInfos.foreach { cis =>
				val traversable = cis.columnToValue(o)
				if (traversable != null) {
					traversable.foreach { nested =>
						val newO = if (isPersisted(nested)) {
							val OneToMany(foreign: TypeRef, foreignColumns: List[Column]) = cis.column
							// update
							val nestedTpe = typeRegistry.typeInfo(nested).tpe
							val keyArgs = nestedTpe.table.toListOfColumnAndValueTuples(nestedTpe.table.primaryKeys, nested)
							driver.doUpdateOneToManyRef(nestedTpe, foreignColumns zip newKeyValues, keyArgs)
							nested
						} else {
							// insert
							insertInner(nested, o, cis.column, newKeyColumnAndValues)
						}
						val cName = cis.column.alias
						addToMap(cName, newO, modifiedTraversables)
					}
				}
			}

			// many to many
			table.manyToManyColumnInfos.foreach { cis =>
				val traversable = cis.columnToValue(o)
				if (traversable != null) {
					traversable.foreach { nested =>
						val nestedTpe = typeRegistry.typeInfo(nested).tpe
						val newO = if (isPersisted(nested)) {
							nested
						} else {
							insert(nested)
						}
						val rightKeyValues = nestedTpe.table.toListOfPrimaryKeyAndValueTuples(newO)
						driver.doInsertManyToMany(nestedTpe, cis.column, newKeyColumnAndValues, rightKeyValues)
						val cName = cis.column.alias
						addToMap(cName, newO, modifiedTraversables)
					}
				}
			}

			val finalMods = modified ++ modifiedTraversables
			tpe.constructor(ValuesMap.fromMutableMap(typeManager, finalMods))
		}

	/**
	 * insert an entity into the database
	 */
	def insert[T](o: T): T =
		{
			insertInner[T, Any](o, null, null, null)
		}
	/**
	 * update an entity
	 */

	private def updateInner[T](o: T, oldValuesMap: ValuesMap, newValuesMap: ValuesMap): T =
		{
			val tpe = typeRegistry.typeInfo(o).tpe
			val table = tpe.table

			val modified = oldValuesMap.toMutableMap ++ newValuesMap.toMutableMap
			val modifiedTraversables = new HashMap[String, List[Any]]

			def onlyChanged(column: ColumnBase) = newValuesMap(column.columnName) != oldValuesMap(column.columnName)

			// first, lets update the simple types that changed
			val columnsChanged = table.simpleTypeNotAutoGeneratedColumns.filter(onlyChanged _)
			if (!columnsChanged.isEmpty) {
				val args = newValuesMap.toListOfColumnAndValueTuple(columnsChanged)
				val pkArgs = oldValuesMap.toListOfColumnAndValueTuple(table.primaryKeys)
				driver.doUpdate(tpe, args, pkArgs)
			}

			// next, update one-to-many
			table.oneToManyColumns.foreach { oneToMany =>
				val t: Traversable[Any] = table.findOneToManyTraversableColumnInfo(oneToMany).columnToValue(o)

				// we'll get the 2 traversables and update the database
				// based on their differences
				val newValues = t.toList
				val oldValues: Seq[Any] = oldValuesMap.seq(oneToMany.foreign.alias)

				// update those that remained in the updated traversable
				val intersection = newValues.intersect(oldValues)
				intersection.foreach { item =>
					val newItem = update(item)
					addToMap(oneToMany.alias, newItem, modifiedTraversables)
				}
				// find the added ones
				val diff = newValues.diff(oldValues)
				diff.foreach { item =>
					val keysAndValues = table.primaryKeys.map(_.column) zip table.primaryKeys.map(c => modified(c.columnName))
					val newItem = insertInner(item, o, oneToMany, keysAndValues);
					addToMap(oneToMany.alias, newItem, modifiedTraversables)
				}

				// find the removed ones
				val odiff = oldValues.diff(newValues)
				odiff.foreach { item =>
					delete(item)
				}
			}
			// now update o.valuesMap
			val finalValuesMap = ValuesMap.fromMutableMap(typeManager, modified ++ modifiedTraversables)
			val r = tpe.constructor(finalValuesMap)
			r
		}

	/**
	 * update a mutable entity. The entity must have been retrieved from the database and then changed prior to calling this method.
	 * The whole tree will be updated (if necessary).
	 * The method heavily relies on object equality to assess which entities will be updated.
	 */
	def update[T](o: T): T =
		{
			if (!o.isInstanceOf[Persisted]) throw new IllegalArgumentException("can't update an object that is not persisted: " + o);
			val persisted = o.asInstanceOf[Persisted]
			if (persisted.discarded) throw new IllegalArgumentException("can't operate on an object twice. An object that was updated/deleted must be discarded and replaced by the return value of update(), i.e. onew=update(o) or just be disposed if it was deleted. The offending object was : " + o);
			persisted.discarded = true
			val oldValuesMap = persisted.valuesMap
			val newValuesMap = ValuesMap.fromEntity(typeManager, typeRegistry.typeInfo(o).tpe, o)
			updateInner(o, oldValuesMap, newValuesMap)
		}

	/**
	 * update an immutable entity. The entity must have been retrieved from the database. Because immutables can't change, a new instance
	 * of the entity must be created with the new values prior to calling this method. Values that didn't change should be copied from o.
	 * The method heavily relies on object equality to assess which entities will be updated.
	 * The whole tree will be updated (if necessary).
	 *
	 * @param	o		the entity, as retrieved from the database
	 * @param	newO	the new instance of the entity with modifications
	 * @return			The updated entity. Both o and newO should be disposed (not used) after the call.
	 */
	def update[T](o: T, newO: T): T =
		{
			if (!o.isInstanceOf[Persisted]) throw new IllegalArgumentException("can't update an object that is not persisted: " + o);
			val persisted = o.asInstanceOf[Persisted]
			if (persisted.discarded) throw new IllegalArgumentException("can't operate on an object twice. An object that was updated/deleted must be discarded and replaced by the return value of update(), i.e. onew=update(o) or just be disposed if it was deleted. The offending object was : " + o);
			persisted.discarded = true
			val oldValuesMap = persisted.valuesMap
			val newValuesMap = ValuesMap.fromEntity(typeManager, typeRegistry.typeInfo(newO).tpe, newO)
			updateInner(newO, oldValuesMap, newValuesMap)
		}
	/**
	 * select an entity by it's ID
	 *
	 * @param clz		Class[T], classOf[Entity]
	 * @param id		the id
	 * @return			Option[T] or None
	 */
	def select[T](clz: Class[T], id: AnyVal): Option[T] = select(clz, List(id))

	def select[T](clz: Class[T], ids: List[AnyVal]): Option[T] =
		{
			val tpe = typeRegistry.typeInfo(clz).tpe
			if (tpe.table.primaryKeys.size != ids.size) throw new IllegalStateException("Primary keys number dont match the number of parameters. Primary keys: %s".format(tpe.table.primaryKeys))

			val om = driver.doSelect(tpe, tpe.table.primaryKeys.map(_.column).zip(ids))
			if (om.isEmpty) None
			else if (om.size > 1) throw new IllegalStateException("expected 1 result for %s and ids %s, but got %d. Is the primary key column a primary key in the table?".format(clz.getSimpleName, ids, om.size))
			else {
				val l = toEntities(om, tpe, new EntityMap)
				if (l.size != 1) throw new IllegalStateException("expected 1 object, but got %s".format(l))
				Option(l.head)
			}
		}

	private def toEntities[T](lm: List[JdbcMap], tpe: Type[T], entities: EntityMap): List[T] = lm.map { om =>
		val mods = new scala.collection.mutable.HashMap[String, Any]
		mods ++= om.map
		val table = tpe.table
		// calculate the id's for this tpe
		val ids: List[Any] = tpe.table.primaryKeys.map { pk => om(pk.column.columnName) }
		val entity = entities.get[T](tpe.clz, ids)
		if (entity.isDefined) {
			entity.get
		} else {
			def createMock: T =
				{
					mods ++= table.oneToManyColumns.map(c => (c.alias -> List()))
					mods ++= table.manyToManyColumns.map(c => (c.alias -> List()))
					// create a mock of the final entity, to avoid cyclic dependencies
					tpe.constructor(ValuesMap.fromMutableMap(typeManager, mods))
				}
			entities.put(tpe.clz, ids, createMock)
			// one to many
			table.oneToManyColumns.foreach { c =>
				val ftpe = typeRegistry.typeInfo(c.foreign.clz).tpe
				val fom = driver.doSelect(ftpe, c.foreignColumns.zip(ids))
				val otmL = toEntities(fom, ftpe, entities)
				mods(c.foreign.alias) = otmL
			}

			// many to many
			table.manyToManyColumns.foreach { c =>
				val ftpe = typeRegistry.typeInfo(c.foreign.clz).tpe
				val fom = driver.doSelectManyToMany(tpe, ftpe, c, c.linkTable.left zip ids)
				val mtmR = toEntities(fom, ftpe, entities)
				mods(c.foreign.alias) = mtmR
			}

			val vm = ValuesMap.fromMutableMap(typeManager, mods)
			val entity = tpe.constructor(vm)
			entities.reput(tpe.clz, ids, entity)
			entity
		}
	}

	/**
	 * deletes an entity from the database
	 */
	def delete[T](o: T): T =
		{
			if (!o.isInstanceOf[Persisted]) throw new IllegalArgumentException("can't delete an object that is not persisted: " + o);

			val persisted = o.asInstanceOf[Persisted]
			if (persisted.discarded) throw new IllegalArgumentException("can't operate on an object twice. An object that was updated/deleted must be discarded and replaced by the return value of update(), i.e. onew=update(o) or just be disposed if it was deleted. The offending object was : " + o);
			persisted.discarded = true

			val tpe = typeRegistry.typeInfo(o).tpe
			val table = tpe.table

			val keyValues = table.toListOfPrimaryKeyAndValueTuples(o)
			driver.doDelete(tpe, keyValues)
			o
		}

	/**
	 * ===================================================================================
	 * ID helper methods
	 * ===================================================================================
	 */
	/**
	 * retrieve the id of an entity
	 */
	def intIdOf(o: AnyRef) = o match {
		case iid: IntId => iid.id
	}

	/**
	 * retrive the id of an entity
	 */
	def longIdOf(o: AnyRef) = o match {
		case iid: LongId => iid.id
	}

	/**
	 * ===================================================================================
	 * common methods
	 * ===================================================================================
	 */
	override def toString = "MapperDao(%s)".format(driver)
}
