package com.rits.orm
/**
 * mapping tables to entities
 *
 * @author kostantinos.kougios
 *
 * 12 Jul 2011
 */

protected abstract class ColumnBase {
	def columnName: String
	def alias: String

	def isSimpleType = this match {
		case _: Column => true
		case _: PK => true
		case _: AutoGenerated => true
		case _ => false
	}

	def isAutoGenerated = this match {
		case _: AutoGenerated => true
		case PK(_: AutoGenerated) => true
		case _ => false
	}
}

case class Table[T](name: String, columnInfos: List[ColumnInfoBase[T, _]]) {
	// the primary keys for this table
	val primaryKeys: List[PK] = columnInfos.map(_.column).collect {
		case pk: PK => pk
	}

	val columns: List[ColumnBase] = columnInfos.map(_.column)
	val columnsWithoutAutoGenerated: List[ColumnBase] = columnInfos.map(_.column).filterNot(_.isAutoGenerated)

	val simpleTypeColumns: List[ColumnBase] = columns.collect {
		case c: Column => c
		case pk: PK => pk
	}
	val simpleTypeAutoGeneratedColumns: List[ColumnBase] = simpleTypeColumns.filter(_.isAutoGenerated)
	val simpleTypeNotAutoGeneratedColumns: List[ColumnBase] = simpleTypeColumns.filterNot(_.isAutoGenerated)

	val simpleTypeColumnInfos = columnInfos.filter(_.isSimpleTypeColumn)

	val oneToManyColumns: List[OneToMany] = columns.collect {
		case c: OneToMany => c
	}
	val manyToManyColumns: List[ManyToMany] = columns.collect {
		case c: ManyToMany => c
	}

	val oneToManyColumnInfos: List[ColumnInfoTraversableOneToMany[T]] = columnInfos.collect {
		case c: ColumnInfoTraversableOneToMany[T] => c
	}

	val manyToManyColumnInfos: List[ColumnInfoTraversableManyToMany[T]] = columnInfos.collect {
		case c: ColumnInfoTraversableManyToMany[T] => c
	}

	val columnToColumnInfoMap: Map[ColumnBase, ColumnInfoBase[T, _]] = columnInfos.map(ci => (ci.column, ci)).toMap

	def findOneToManyTraversableColumnInfo(column: ColumnBase): ColumnInfoTraversableOneToMany[T] =
		{
			val l = columnInfos.collect {
				case c: ColumnInfoTraversableOneToMany[T] if (c.column == column) => c
			}
			require(l.size == 1)
			l.head
		}

	def toListOfPrimaryKeyValues(o: T): List[Any] = primaryKeys.map(c => columnToColumnInfoMap(c)).map(_.columnToValue(o))
	def toListOfPrimaryKeyAndValueTuples(o: T): List[(PK, Any)] = primaryKeys.map { c => (c, columnToColumnInfoMap(c).columnToValue(o)) }

	def toListOfColumnAndValueTuples(columns: List[ColumnBase], o: T): List[(ColumnBase, Any)] = columns.map { c => (c, columnToColumnInfoMap(c).columnToValue(o)) }

	def toColumnAndValueMap(columns: List[ColumnBase], o: T): Map[ColumnBase, Any] = columns.map { c => (c, columnToColumnInfoMap(c).columnToValue(o)) }.toMap

	def toColumnAliasAndValueMap(columns: List[ColumnBase], o: T): Map[String, Any] = toColumnAndValueMap(columns, o).map(e => (e._1.alias, e._2))
}

case class LinkTable(name: String, left: List[Column], right: List[Column])
object LinkTable {
	def apply(name: String, left: String, right: String): LinkTable = LinkTable(name, Column.many(left), Column.many(right))
}

/**
 * Columns
 */
abstract class SimpleColumn extends ColumnBase

case class Column(name: String) extends SimpleColumn {
	def columnName = name
	def alias = name
}
object Column {
	def many(name: String): List[Column] = List(Column(name))
	def many(name1: String, name2: String): List[Column] = List(Column(name1), Column(name2))
}

case class AutoGenerated(name: String) extends SimpleColumn {
	def columnName = name
	def alias = name
}
case class PK(column: SimpleColumn) extends ColumnBase {
	def columnName = column.columnName
	def alias = columnName
}
case class OneToMany(foreign: TypeRef, foreignColumns: List[Column]) extends ColumnBase {
	def columnName = throw new IllegalStateException("OneToMany doesn't have a columnName")
	def alias = foreign.alias
}
case class ManyToOne(column: Column, foreign: TypeRef) extends ColumnBase {
	def columnName = column.name
	def alias = column.name
}
case class ManyToMany(linkTable: LinkTable, foreign: TypeRef) extends ColumnBase {
	def columnName = throw new IllegalStateException("ManyToMany doesn't have a columnName")
	def alias = foreign.alias
}

case class Type[T](private val typeClass: Class[_], constructor: ValuesMap => T with Persisted, table: Table[T]) {
	def clz: Class[T] = typeClass.asInstanceOf[Class[T]]
}
case class TypeRef(alias: String, clz: Class[_])

/**
 * Column Infos
 */
class ColumnInfoBase[T, R](val column: ColumnBase, val columnToValue: T => R) {
	def isSimpleTypeColumn = column.isSimpleType
}
case class ColumnInfo[T](override val column: ColumnBase, override val columnToValue: T => Any) extends ColumnInfoBase[T, Any](column, columnToValue)

case class ColumnInfoTraversableOneToMany[T](override val column: OneToMany, override val columnToValue: T => Traversable[Any]) extends ColumnInfoBase[T, Traversable[Any]](column, columnToValue)

case class ColumnInfoTraversableManyToMany[T](override val column: ManyToMany, override val columnToValue: T => Traversable[Any]) extends ColumnInfoBase[T, Traversable[Any]](column, columnToValue)

object Mappings {
	/**
	 * one to many mapping from T to a traversable of an other entity E
	 *
	 * @alias			the alias that can be used to construct T.
	 * @foreignClz		the refered entity
	 * @columnToValue	the function from T => Traversable[Any] that accesses the field of T to return the Traversable of E
	 */
	def oneToMany[T](alias: String, foreignClz: Class[_], foreignKeyColumn: String, columnToValue: T => Traversable[Any]): ColumnInfoTraversableOneToMany[T] =
		{
			ColumnInfoTraversableOneToMany[T](OneToMany(TypeRef(alias, foreignClz), Column.many(foreignKeyColumn)), columnToValue)
		}

	def manyToMany[T](alias: String, linkTable: String, leftColumn: String, rightColumn: String, referencedType: Class[_], columnToValue: T => Traversable[Any]): ColumnInfoTraversableManyToMany[T] =
		{
			ColumnInfoTraversableManyToMany[T](
				ManyToMany(
					LinkTable(linkTable, List(Column(leftColumn)), List(Column(rightColumn))),
					TypeRef(alias, referencedType)
				),
				columnToValue
			)
		}
	/**
	 * basic mapping to a database column
	 */
	def basic[T](column: String, columnToValue: T => Any): ColumnInfo[T] = ColumnInfo(Column(column), columnToValue)
	/**
	 * an autogenerated column that is a primary key
	 */
	def autoGeneratedPK[T](idColumn: String, columnToValue: T => Any): ColumnInfo[T] = ColumnInfo(PK(AutoGenerated(idColumn)), columnToValue)
	/**
	 * map a primary key column. Use autoGeneratedPK to map a primary key that is auto generated
	 */
	def pk[T](idColumn: String, columnToValue: T => Any): ColumnInfo[T] = ColumnInfo(PK(Column(idColumn)), columnToValue)
	/**
	 * the main entry to create a mapping, by passing the class, the constructor function, the table name and the colums.
	 * Columns can be constructed using the methods basic,oneToMany etc.
	 */
	def entity[T](clz: Class[_], constructor: ValuesMap => T with Persisted, tableName: String, columns: ColumnInfoBase[T, _]*): Type[T] =
		{
			Type[T](clz, constructor, Table(tableName, columns.toList))
		}
}
