package com.rits.orm.drivers
import com.rits.orm._
import com.rits.jdbc.Jdbc
import com.rits.jdbc.JdbcMap
import com.rits.jdbc.UpdateResultWithGeneratedKeys
/**
 * all database drivers must implement this trait
 *
 * @author kostantinos.kougios
 *
 * 14 Jul 2011
 */
trait Driver {
	val jdbc: Jdbc
	/**
	 * =====================================================================================
	 * utility methods
	 * =====================================================================================
	 */
	protected def escapeColumnNames(name: String): String = name

	protected def commaSeparatedListOfSimpleTypeColumns[T](separator: String, columns: List[ColumnBase]): String = columns.map(_.columnName).map(escapeColumnNames _).mkString(separator)

	protected def generateColumnsEqualsValueString(l: List[ColumnBase]): String =
		{
			val sb = new StringBuilder(20)
			var cnt = 0
			l.foreach { ci =>
				if (cnt > 0) sb.append(",\n") else cnt += 1
				sb append escapeColumnNames(ci.columnName) append "=?"
			}
			sb.toString
		}
	protected def generateColumnsEqualsValueString(prefix: String, separator: String, l: List[ColumnBase]): String =
		{
			val sb = new StringBuilder(20)
			var cnt = 0
			l.foreach { ci =>
				if (cnt > 0) sb.append(separator) else cnt += 1
				sb append prefix append escapeColumnNames(ci.columnName) append "=?"
			}
			sb.toString
		}

	/**
	 * =====================================================================================
	 * INSERT
	 * =====================================================================================
	 */

	/**
	 * default implementation of insert, should do for most subclasses
	 */
	def doInsert[T](tpe: Type[T], args: List[(ColumnBase, Any)]): UpdateResultWithGeneratedKeys =
		{
			val filteredArgs = args.filter { cv =>
				cv._1 match {
					case _: AutoGenerated => false
					case PK(_: AutoGenerated) => false
					case _ => true
				}
			}
			val sql = insertSql(tpe, filteredArgs)
			jdbc.updateGetAutoGenerated(sql, filteredArgs.map(_._2))
		}

	/**
	 * default impl of the insert statement generation
	 */
	protected def insertSql[T](tpe: Type[T], args: List[(ColumnBase, Any)]): String =
		{
			val sb = new StringBuilder(100, "insert into ")
			sb append tpe.table.name append "(" append commaSeparatedListOfSimpleTypeColumns(",", args.map(_._1))
			sb append ")\n"
			sb append "values(?" append (",?" * (args.size - 1)) append ")"

			sb.toString
		}

	def doInsertManyToMany[T](tpe: Type[T], manyToMany: ManyToMany, left: List[(ColumnBase, Any)], right: List[(ColumnBase, Any)]): Unit =
		{
			val sql = insertManyToManySql(tpe, manyToMany, left, right)
			jdbc.update(sql, left.map(_._2) ::: right.map(_._2))
		}

	protected def insertManyToManySql[T](tpe: Type[T], manyToMany: ManyToMany, left: List[(ColumnBase, Any)], right: List[(ColumnBase, Any)]): String =
		{
			val sb = new StringBuilder(100, "insert into ")
			val linkTable = manyToMany.linkTable
			sb append linkTable.name append "(" append commaSeparatedListOfSimpleTypeColumns(",", linkTable.left) append "," append commaSeparatedListOfSimpleTypeColumns(",", linkTable.right) append ")\n"
			sb append "values(?" append (",?" * (linkTable.left.size - 1 + linkTable.right.size)) append ")"
			sb.toString
		}
	/**
	 * =====================================================================================
	 * UPDATE
	 * =====================================================================================
	 */
	/**
	 * default implementation of update, should do for most subclasses
	 */
	def doUpdate[T](tpe: Type[T], args: List[(ColumnBase, Any)], pkArgs: List[(ColumnBase, Any)]): UpdateResultWithGeneratedKeys =
		{
			val sql = updateSql(tpe, args, pkArgs)
			jdbc.updateGetAutoGenerated(sql, args.map(_._2) ::: pkArgs.map(_._2))
		}
	/**
	 * default impl of the insert statement generation
	 */
	protected def updateSql[T](tpe: Type[T], args: List[(ColumnBase, Any)], pkArgs: List[(ColumnBase, Any)]): String =
		{
			val sb = new StringBuilder(100, "update ")
			sb append tpe.table.name append "\n"
			sb append "set " append generateColumnsEqualsValueString(args.map(_._1))
			sb append "\nwhere " append generateColumnsEqualsValueString(pkArgs.map(_._1))
			sb.toString
		}

	/**
	 * links one-to-many objects to their parent
	 */
	def doUpdateOneToManyRef[T](tpe: Type[T], foreignKeys: List[(ColumnBase, Any)], pkArgs: List[(ColumnBase, Any)]) =
		{
			val sql = updateOneToManyRefSql(tpe, foreignKeys, pkArgs)
			jdbc.update(sql, foreignKeys.map(_._2) ::: pkArgs.map(_._2))
		}

	protected def updateOneToManyRefSql[T](tpe: Type[T], foreignKeys: List[(ColumnBase, Any)], pkArgs: List[(ColumnBase, Any)]): String =
		{
			val sb = new StringBuilder(100, "update ")
			sb append tpe.table.name append "\n"
			sb append "set " append generateColumnsEqualsValueString(foreignKeys.map(_._1))
			sb append "\nwhere " append generateColumnsEqualsValueString(pkArgs.map(_._1))
			sb.toString
		}
	/**
	 * =====================================================================================
	 * SELECT
	 * =====================================================================================
	 */

	/**
	 * default impl of select
	 */
	def doSelect[T](tpe: Type[T], where: List[(SimpleColumn, Any)]): List[JdbcMap] =
		{
			val sql = selectSql(tpe, where)

			// 1st step is to get the simple values
			// of this object from the database
			jdbc.queryForList(sql, where.map(_._2))
		}

	protected def selectSql[T](tpe: Type[T], where: List[(SimpleColumn, Any)]): String =
		{
			val columns = tpe.table.simpleTypeColumns
			val sb = new StringBuilder(100, "select ")
			sb append commaSeparatedListOfSimpleTypeColumns(",", columns) append " from " append tpe.table.name
			sb append "\nwhere " append generateColumnsEqualsValueString(where.map(_._1))

			sb.toString
		}

	def doSelectManyToMany[T, F](tpe: Type[T], ftpe: Type[F], manyToMany: ManyToMany, leftKeyValues: List[(SimpleColumn, Any)]): List[JdbcMap] =
		{
			val sql = selectManyToManySql(tpe, ftpe, manyToMany, leftKeyValues)
			jdbc.queryForList(sql, leftKeyValues.map(_._2))
		}

	protected def selectManyToManySql[T, F](tpe: Type[T], ftpe: Type[F], manyToMany: ManyToMany, leftKeyValues: List[(SimpleColumn, Any)]): String =
		{
			val ftable = ftpe.table
			val columns = ftable.simpleTypeColumns
			val linkTable = manyToMany.linkTable
			val sb = new StringBuilder(100, "select ")
			sb append commaSeparatedListOfSimpleTypeColumns(",", columns) append "\nfrom " append ftpe.table.name append " f\n"
			sb append "inner join " append linkTable.name append " l on "
			var i = 0
			ftable.primaryKeys.zip(linkTable.right).foreach { z =>
				val PK(left) = z._1
				val right = z._2
				if (i > 0) sb append " and "
				sb append "f." append left.columnName append "=l." append right.name
				i += 1
			}
			sb append "\nwhere " append generateColumnsEqualsValueString("l.", " and ", leftKeyValues.map(_._1))
			sb.toString
		}
	/**
	 * =====================================================================================
	 * DELETE
	 * =====================================================================================
	 */
	def doDelete[T](tpe: Type[T], keyValues: List[(PK, Any)]): Unit =
		{
			val sql = deleteSql(tpe, keyValues)
			jdbc.update(sql, keyValues.map(_._2))
		}

	protected def deleteSql[T](tpe: Type[T], keyValues: List[(PK, Any)]): String =
		{
			val sb = new StringBuilder(100, "delete from ")
			sb append tpe.table.name append " where " append generateColumnsEqualsValueString(keyValues.map(_._1))

			sb.toString
		}

	/**
	 * =====================================================================================
	 * standard methods
	 * =====================================================================================
	 */
	override def toString = "Driver(%s)".format(jdbc)
}