package com.rits.orm

import org.specs2.mutable.SpecificationWithJUnit
import com.rits.jdbc.Jdbc
import com.rits.jdbc.Setup

/**
 * @author kostantinos.kougios
 *
 * 6 Aug 2011
 */
class ManyToManyAutoGeneratedSpec extends SpecificationWithJUnit {

	import ManyToManyAutoGeneratedSpec._

	val (jdbc, mapperDao) = setup

	"insert tree of entities" in {
		createTables
		val product = Product("blue jean", Set(Attribute("colour", "blue", Set()), Attribute("size", "medium", Set())))
		val inserted = mapperDao.insert(product)
		inserted must_== product

		// due to cyclic reference, the attributes set contains "mock" products which have empty traversables.
		// it is not possible to create cyclic-depended immutable instances.
		mapperDao.select(classOf[Product], mapperDao.intIdOf(inserted)).get must_== Product("blue jean",
			Set(
				Attribute("colour", "blue", Set(Product("blue jean", Set()))),
				Attribute("size", "medium", Set(Product("blue jean", Set())))
			)
		)

		// attributes->product should also work
		val colour = inserted.attributes.toList.filter(_.name == "colour").head
		val loadedAttribute = mapperDao.select(classOf[Attribute], mapperDao.intIdOf(colour)).get
		loadedAttribute must_== Attribute("colour", "blue", Set(
			Product("blue jean",
				Set(
					// these are partially loaded due to cyclic dependencies
					Attribute("colour", "blue", Set()),
					Attribute("size", "medium", Set(Product("blue jean", Set())))
				)
			)
		)
		)
	}

	"insert tree of entities with persisted leaf entities" in {
		createTables
		val a1 = mapperDao.insert(Attribute("colour", "blue", Set()))
		val a2 = mapperDao.insert(Attribute("size", "medium", Set()))
		val product = Product("blue jean", Set(a1, a2))
		val inserted = mapperDao.insert(product)
		inserted must_== product

		// due to cyclic reference, the attributes collection contains "mock" products which have empty traversables
		mapperDao.select(classOf[Product], mapperDao.intIdOf(inserted)).get must_== Product("blue jean",
			Set(
				Attribute("colour", "blue", Set(Product("blue jean", Set()))),
				Attribute("size", "medium", Set(Product("blue jean", Set())))
			)
		)
	}

	def createTables =
		{
			jdbc.update("drop table if exists Product_Attribute cascade")
			jdbc.update("drop table if exists Product cascade")
			jdbc.update("drop table if exists Attribute cascade")

			jdbc.update("""
					create table Product (
						id serial not null,
						name varchar(100) not null,
						primary key(id)
					)
			""")
			jdbc.update("""
					create table Attribute (
						id serial not null,
						name varchar(100) not null,
						value varchar(100) not null,
						primary key(id)
					)
			""")
			jdbc.update("""
					create table Product_Attribute (
						product_id int not null,
						attribute_id int not null,
						primary key(product_id,attribute_id)
					)
			""")
		}
	def setup =
		{
			val typeRegistry = TypeRegistry(product, attribute)

			Setup.setupMapperDao(typeRegistry)
		}

}

object ManyToManyAutoGeneratedSpec {
	case class Product(val name: String, val attributes: Set[Attribute])
	case class Attribute(val name: String, val value: String, val products: Set[Product])

	import Mappings._
	val product = entity[Product with IntId](
		classOf[Product],
		m => new Product(
			m.string("name"),
			m.set("attributesAlias")
		) with Persisted with IntId {
			def valuesMap = m
			val id = m.int("id")
		},
		"Product",
		autoGeneratedPK("id", _.id),
		basic("name", _.name),
		manyToMany("attributesAlias", "Product_Attribute", "product_id", "attribute_id", classOf[Attribute], _.attributes)
	)

	val attribute = entity[Attribute with IntId](
		classOf[Attribute],
		m => new Attribute(
			m.string("name"),
			m.string("value"),
			m.set("productsAlias")
		) with Persisted with IntId {
			def valuesMap = m
			val id = m.int("id")
		},
		"Attribute",
		autoGeneratedPK("id", _.id),
		basic("name", _.name),
		basic("value", _.value),
		manyToMany("productsAlias", "Product_Attribute", "attribute_id", "product_id", classOf[Product], _.products)
	)
}