package com.rits.orm
import org.specs2.mutable.SpecificationWithJUnit
import com.rits.jdbc.Jdbc
import com.rits.jdbc.Setup

/**
 * @author kostantinos.kougios
 *
 * 8 Aug 2011
 */
class ManyToManySpec extends SpecificationWithJUnit {
	import ManyToManySpec._

	val (jdbc, mapperDao) = setup

	"insert tree of entities" in {
		createTables
		val product = Product(5, "blue jean", Set(Attribute(2, "colour", "blue"), Attribute(7, "size", "medium")))
		val inserted = mapperDao.insert(product)
		inserted must_== product

		// due to cyclic reference, the attributes set contains "mock" products which have empty traversables.
		// it is not possible to create cyclic-depended immutable instances.
		mapperDao.select(classOf[Product], 5).get must_== inserted

		// attributes->product should also work
		mapperDao.select(classOf[Attribute], 2).get must_== Attribute(2, "colour", "blue")
		mapperDao.select(classOf[Attribute], 2).get must_== Attribute(7, "size", "medium")
	}

	"insert tree of entities with persisted leaf entities" in {
		createTables
		val a1 = mapperDao.insert(Attribute(6, "colour", "blue"))
		val a2 = mapperDao.insert(Attribute(9, "size", "medium"))
		val product = Product(2, "blue jean", Set(a1, a2))
		val inserted = mapperDao.insert(product)
		inserted must_== product

		// due to cyclic reference, the attributes collection contains "mock" products which have empty traversables
		mapperDao.select(classOf[Product], 2).get must_== inserted
	}

	def createTables =
		{
			jdbc.update("drop table if exists Product_Attribute cascade")
			jdbc.update("drop table if exists Product cascade")
			jdbc.update("drop table if exists Attribute cascade")

			jdbc.update("""
					create table Product (
						id int not null,
						name varchar(100) not null,
						primary key(id)
					)
			""")
			jdbc.update("""
					create table Attribute (
						id int not null,
						name varchar(100) not null,
						value varchar(100) not null,
						primary key(id)
					)
			""")
			jdbc.update("""
					create table Product_Attribute (
						product_id int not null,
						attribute_id int not null,
						primary key(product_id,attribute_id)
					)
			""")
		}
	def setup =
		{
			val typeRegistry = TypeRegistry(product, attribute)

			Setup.setupMapperDao(typeRegistry)
		}
}

object ManyToManySpec {
	case class Product(val id: Int, val name: String, val attributes: Set[Attribute])
	case class Attribute(val id: Int, val name: String, val value: String)

	import Mappings._
	val product = entity[Product](
		classOf[Product],
		m => new Product(
			m.int("id"),
			m.string("name"),
			m.set("attributesAlias")
		) with Persisted {
			def valuesMap = m
		},
		"Product",
		pk("id", _.id),
		basic("name", _.name),
		manyToMany("attributesAlias", "Product_Attribute", "product_id", "attribute_id", classOf[Attribute], _.attributes)
	)

	val attribute = entity[Attribute](
		classOf[Attribute],
		m => new Attribute(
			m.int("id"),
			m.string("name"),
			m.string("value")
		) with Persisted {
			def valuesMap = m
		},
		"Attribute",
		pk("id", _.id),
		basic("name", _.name),
		basic("value", _.value)
	)
}