package com.rits.orm

import org.specs2.mutable.SpecificationWithJUnit
import com.rits.jdbc.Jdbc
import com.rits.jdbc.Setup

/**
 * this spec is self contained, all entities, mapping are contained in this class
 *
 * @author kostantinos.kougios
 *
 * 12 Jul 2011
 */
class OneToManyAutoGeneratedSpec extends SpecificationWithJUnit {

	import OneToManyAutoGeneratedSpec._

	val (jdbc, mapperDao) = setup

	"updating items (immutable)" in {
		createTables

		val jp1 = new JobPosition("C++ Developer", 10)
		val jp2 = new JobPosition("Scala Developer", 10)
		val jp3 = new JobPosition("Java Developer", 10)
		val jp4 = new JobPosition("Web Designer", 10)
		val jp5 = new JobPosition("Graphics Designer", 10)

		// we need to apply the same sorting for JobPositions in order for the case classes to be equal because order of list items matters
		val person = Person("Kostas", "K", Set(House("London"), House("Rhodes")), 16, List(jp1, jp2, jp3).sortWith(_.name < _.name))
		val inserted = mapperDao.insert(person)
		inserted must_== person
		var updated: Person = inserted
		def doUpdate(from: Person, to: Person) =
			{
				updated = mapperDao.update(from, to)
				val localUpdated = updated // for debugging
				updated must_== to
				val id = mapperDao.intIdOf(updated)
				val loaded = mapperDao.select(classOf[Person], id).get
				loaded must_== updated
				loaded must_== to
			}
		doUpdate(updated, new Person("Changed", "K", updated.owns, 18, updated.positions.filterNot(_ == jp1)))
		doUpdate(updated, new Person("Changed Again", "Surname changed too", updated.owns.filter(_.address == "London"), 18, jp5 :: updated.positions.filterNot(jp => jp == jp1 || jp == jp3)))
	}

	"updating items (mutable)" in {
		createTables

		val jp1 = new JobPosition("C++ Developer", 10)
		val jp2 = new JobPosition("Scala Developer", 10)
		val jp3 = new JobPosition("Java Developer", 10)
		val person = new Person("Kostas", "K", Set(House("London"), House("Rhodes")), 16, List(jp1, jp2, jp3))
		val inserted = mapperDao.insert(person)

		inserted.positions.foreach(_.name = "changed")
		inserted.positions.foreach(_.rank = 5)
		val updated = mapperDao.update(inserted)
		updated must_== inserted

		val loaded = mapperDao.select(classOf[Person], mapperDao.intIdOf(inserted)).get
		loaded must_== updated
	}

	"removing items" in {
		createTables

		val jp1 = new JobPosition("C++ Developer", 10)
		val jp2 = new JobPosition("Scala Developer", 10)
		val jp3 = new JobPosition("Java Developer", 10)
		val person = new Person("Kostas", "K", Set(House("London"), House("Rhodes")), 16, List(jp1, jp2, jp3))
		val inserted = mapperDao.insert(person)

		inserted.positions = inserted.positions.filterNot(jp => jp == jp1 || jp == jp3)
		val updated = mapperDao.update(inserted)
		updated must_== inserted

		val loaded = mapperDao.select(classOf[Person], mapperDao.intIdOf(inserted)).get
		loaded must_== updated
	}

	"adding items" in {
		createTables

		val person = new Person("Kostas", "K", Set(House("London"), House("Rhodes")), 16, List(new JobPosition("Scala Developer", 10), new JobPosition("Java Developer", 10)))
		val inserted = mapperDao.insert(person)
		val id = mapperDao.intIdOf(inserted)

		val loaded = mapperDao.select(classOf[Person], id).get

		// add more elements to the collection
		loaded.positions = (new JobPosition("Groovy Developer", 5) :: new JobPosition("C++ Developer", 8) :: loaded.positions).sortWith(_.name < _.name)
		val updatedPositions = mapperDao.update(loaded)
		updatedPositions must_== loaded

		val updatedReloaded = mapperDao.select(classOf[Person], id).get
		updatedReloaded must_== updatedPositions
	}

	"using already persisted entities" in {
		createTables
		jdbc.update("alter table House alter person_id drop not null")
		val h1 = mapperDao.insert(House("London"))
		val h2 = mapperDao.insert(House("Rhodes"))
		val h3 = mapperDao.insert(House("Santorini"))

		val person = new Person("Kostas", "K", Set(h1, h2), 16, List(new JobPosition("Scala Developer", 10), new JobPosition("Java Developer", 10)))
		val inserted = mapperDao.insert(person)
		val id = mapperDao.intIdOf(inserted)

		mapperDao.select(classOf[Person], id).get must_== inserted

		val changed = Person("changed", "K", inserted.owns.filterNot(_ == h1), 18, List())
		val updated = mapperDao.update(inserted, changed)
		updated must_== changed

		mapperDao.select(classOf[Person], id).get must_== updated
	}

	"CRUD (multi purpose test)" in {
		createTables

		val items = for (i <- 1 to 10) yield {
			val person = new Person("Kostas", "K", Set(House("London"), House("Rhodes")), 16, List(new JobPosition("Java Developer", 10), new JobPosition("Scala Developer", 10)))
			val inserted = mapperDao.insert(person)
			inserted must_== person

			val id = mapperDao.intIdOf(inserted)

			val loaded = mapperDao.select(classOf[Person], id).get
			loaded must_== person

			// update

			loaded.name = "Changed"
			loaded.age = 24
			loaded.positions.head.name = "Java/Scala Developer"
			loaded.positions.head.rank = 123
			val updated = mapperDao.update(loaded)
			updated must_== loaded

			val reloaded = mapperDao.select(classOf[Person], id).get
			reloaded must_== loaded

			// add more elements to the collection
			reloaded.positions = new JobPosition("C++ Developer", 8) :: reloaded.positions
			val updatedPositions = mapperDao.update(reloaded)
			updatedPositions must_== reloaded

			val updatedReloaded = mapperDao.select(classOf[Person], id).get
			updatedReloaded must_== updatedPositions

			// remove elements from the collection
			updatedReloaded.positions = updatedReloaded.positions.filterNot(_ == updatedReloaded.positions(1))
			val removed = mapperDao.update(updatedReloaded)
			removed must_== updatedReloaded

			val removedReloaded = mapperDao.select(classOf[Person], id).get
			removedReloaded must_== removed

			// remove them all
			removedReloaded.positions = List()
			mapperDao.update(removedReloaded) must_== removedReloaded

			val rereloaded = mapperDao.select(classOf[Person], id).get
			rereloaded must_== removedReloaded
			rereloaded.positions = List(JobPosition("Java Developer %d".format(i), 15 + i))
			rereloaded.name = "final name %d".format(i)
			val reupdated = mapperDao.update(rereloaded)

			val finalLoaded = mapperDao.select(classOf[Person], id).get
			finalLoaded must_== reupdated
			(id, finalLoaded)
		}
		val ids = items.map(_._1)
		ids.toSet.size must_== ids.size
		// verify all still available
		items.foreach { item =>
			mapperDao.select(classOf[Person], item._1).get must_== item._2
		}
		success
	}

	def createTables {
		jdbc.update("drop table if exists JobPosition cascade")
		jdbc.update("""
			create table JobPosition (
				id serial not null,
				name varchar(100) not null,
				start timestamp with time zone,
				"end" timestamp with time zone,
				rank int not null,
				person_id int not null,
				primary key (id),
				constraint FK_JobPosition_Person foreign key (person_id) references Person(id) on delete cascade on update cascade
			)
		""")
		jdbc.update("drop table if exists House cascade")
		jdbc.update("""
			create table House (
				id serial not null,
				address varchar(100) not null,
				person_id int not null,
				primary key (id),
				constraint FK_House_Person foreign key (person_id) references Person(id) on delete cascade on update cascade
			)
		""")
		jdbc.update("drop table if exists Person cascade")
		jdbc.update("""
			create table Person (
				id serial not null,
				name varchar(100) not null,
				surname varchar(100) not null,
				age int not null,
				primary key (id)
			)
		""")
	}
	def setup =
		{
			val typeRegistry = TypeRegistry(jobPosition, house, person)

			Setup.setupMapperDao(typeRegistry)
		}

}

object OneToManyAutoGeneratedSpec {
	/**
	 * ============================================================================================================
	 * the entities
	 * ============================================================================================================
	 */
	/**
	 * the only reason this is a case class, is to ease testing. There is no requirement
	 * for persisted classes to follow any convention.
	 *
	 * Also the only reason for this class to be mutable is for testing. In a real application
	 * it could be immutable.
	 */
	case class JobPosition(var name: String, var rank: Int) {
		// this can have any arbitrary methods, no problem!
		def whatRank = rank
		// also any non persisted fields, no prob! It's up to the mappings regarding which fields will be used
		val whatever = 5
	}

	/**
	 * the only reason this is a case class, is to ease testing. There is no requirement
	 * for persisted classes to follow any convention
	 *
	 * Also the only reason for this class to be mutable is for testing. In a real application
	 * it could be immutable.
	 */
	case class Person(var name: String, val surname: String, owns: Set[House], var age: Int, var positions: List[JobPosition]) {
	}

	case class House(val address: String)

	/**
	 * ============================================================================================================
	 * Mapping for JobPosition class
	 * ============================================================================================================
	 */
	import Mappings._

	val jobPosition = entity[JobPosition with IntId](
		// the class of the mapped object
		classOf[JobPosition],
		// a function from ValuesMap=>JobPosition that constructs the object.
		// This means that immutability is possible and even desirable for entities!
		// Note: m("id"), m("name") etc could be used, but for extra type safety the more specialized methods are.
		m => new JobPosition(m.string("name"), m.int("rank")) with Persisted with IntId {
			// this holds the original values of the object as retrieved from the database.
			// later on it is used to compare what changed in this object.
			def valuesMap = m
			val id = m.int("id")
		},
		// now a description of the table and it's columns follows.
		// each column is followed by a function JobPosition=>Any, that
		// returns the value of the property for that column.
		"JobPosition",
		autoGeneratedPK("id", _.id), // this is the primary key
		basic("name", _.name), // _.name : JobPosition => Any . Function that maps the column to the value of the object
		basic("rank", _.rank)
	)

	val house = entity[House with IntId](
		classOf[House],
		m => new House(m.string("address")) with Persisted with IntId {
			def valuesMap = m
			val id = m.int("id")
		},
		"House",
		autoGeneratedPK("id", _.id),
		basic("address", _.address)
	)
	// mapping for Person class
	val person = entity[Person with IntId](
		classOf[Person],
		m => new Person(m.string("name"), m.string("surname"), m.set[House]("housesAlias"), m.int("age"), m.list[JobPosition]("jobPositionsAlias").sortWith(_.name < _.name)) with Persisted with IntId {
			def valuesMap = m
			val id = m.int("id")
		},
		"Person",
		autoGeneratedPK("id", _.id),
		basic("name", _.name),
		basic("surname", _.surname),
		oneToMany("housesAlias", classOf[House], "person_id", _.owns),
		basic("age", _.age),
		/**
		 * a traversable one-to-many relationship with JobPositions.
		 * The type of the relationship is classOf[JobPosition] and the alias
		 * for retrieving the Traversable is jobPositionsAlias. This is used above, when
		 * creating Person: new Person(....,m.toList("jobPositionsAlias")) .
		 * JobPositions table has a person_id foreign key which references Person table.
		 */
		oneToMany("jobPositionsAlias", classOf[JobPosition], "person_id", _.positions)
	)
}